#include <cmath>
#include "linespacefilter.h"
#include "blurredsegmentproto.h"



const int LineSpaceFilter::DEFAULT_NB_ANG_CELLS = 32;
const int LineSpaceFilter::DEFAULT_NB_DIST_CELLS = 32;
const int LineSpaceFilter::MIN_NB_ANG_CELLS = 8;
const int LineSpaceFilter::MAX_NB_ANG_CELLS = 64;
const int LineSpaceFilter::MIN_NB_DIST_CELLS = 8;
const int LineSpaceFilter::MAX_NB_DIST_CELLS = 64;
const int LineSpaceFilter::DEFAULT_SELECTION_THRESHOLD = 75;
const int LineSpaceFilter::MIN_SELECTION_THRESHOLD = 20;
const int LineSpaceFilter::MAX_SELECTION_THRESHOLD = 90;
const int LineSpaceFilter::INC_SELECTION_THRESHOLD = 5;
const int LineSpaceFilter::NO_SUBPIXEL = 1;
const int LineSpaceFilter::MAX_SUBPIXEL = 64;
const int LineSpaceFilter::ANG_DEF_ZOOM = 8;
const int LineSpaceFilter::ANG_MIN_ZOOM = 2;
const int LineSpaceFilter::ANG_MAX_ZOOM = 32;
const int LineSpaceFilter::DIST_DEF_ZOOM = 4;
const int LineSpaceFilter::DIST_MIN_ZOOM = 1;
const int LineSpaceFilter::DIST_MAX_ZOOM = 16;


LineSpaceFilter::LineSpaceFilter ()
{
  nt = DEFAULT_NB_ANG_CELLS;
  nr = DEFAULT_NB_DIST_CELLS;
  acc = new int[nt*nr];
  ncos = new double[nt];
  nsin = new double[nt];
  mask = new bool[nt*nr];
  cellMax = -1;
  minsel = DEFAULT_SELECTION_THRESHOLD;
  azoom = ANG_DEF_ZOOM;
  dzoom = DIST_DEF_ZOOM;
  nbSubpix = NO_SUBPIXEL;
}


LineSpaceFilter::~LineSpaceFilter ()
{
  delete [] acc;
  delete [] ncos;
  delete [] nsin;
  delete [] mask;
}


void LineSpaceFilter::clear ()
{
  leftIn.clear ();
  rightIn.clear ();
  out.clear ();
}


BlurredSegment *LineSpaceFilter::filter (BlurredSegment *bs)
{
  bsInitialSize = 0;
  bsFinalSize = 0;
  if (bs == NULL || ! bs->isThick ()) return NULL;
  bsInitialSize = bs->size ();
  if (bsInitialSize < 10) return NULL;

  // Gets the line spacefilter
  const Pt2i bsstart = bs->getCenter ();
  const vector<Pt2i> *bslineleft = bs->getLeftLine ();
  const vector<Pt2i> *bsleft = bs->getLeftPoints ();
  const vector<Pt2i> *bslineright = bs->getRightLine ();
  const vector<Pt2i> *bsright = bs->getRightPoints ();
  Vr2i sv = bs->getSupportVector ();
  AbsRat segwidth = bs->segmentRationalWidth ();
  Vr2i bbs = bs->boundingBoxSize ();
  init (bsstart.x (), bsstart.y (), -sv.y (), sv.x (),
        sqrt (bbs.x () * bbs.x () + bbs.y () * bbs.y ()),
        segwidth.numerator () / (double) segwidth.denominator ());
  vote (bsstart);
  for (vector<Pt2i>::const_iterator it = bslineleft->begin ();
       it != bslineleft->end (); it ++)
    vote (*it);
  for (vector<Pt2i>::const_iterator it = bsleft->begin ();
       it != bsleft->end (); it ++)
    vote (*it);
  for (vector<Pt2i>::const_iterator it = bslineright->begin ();
       it != bslineright->end (); it ++)
    vote (*it);
  for (vector<Pt2i>::const_iterator it = bsright->begin ();
       it != bsright->end (); it ++)
    vote (*it);
  findMaxThroughOrigin ();
  fitMask ();

  // Filters the previous points
  for (vector<Pt2i>::const_iterator it = bsleft->begin ();
       it != bsleft->end (); it ++)
  {
    if (voteInMask (*it)) leftIn.push_back (Pt2i (*it));
    else out.push_back (Pt2i (*it));
  }
  for (vector<Pt2i>::const_iterator it = bsright->begin ();
       it != bsright->end (); it ++)
  {
    if (voteInMask (*it)) rightIn.push_back (Pt2i (*it));
    else out.push_back (Pt2i (*it));
  }

  // Returns a blurred segment with the accepted points
  int mw = 1;
  AbsRat sw = bs->segmentRationalWidth ();
  if (sw.denominator () != 0) mw += sw.numerator () / sw.denominator ();
  BlurredSegmentProto resbs (mw, bsstart, leftIn, rightIn);
  bsFinalSize = resbs.size ();
  return (resbs.endOfBirth ());
}


void LineSpaceFilter::getAccumulator (unsigned char *map, int *solution)
{
  if (cellMax != -1)
  {
    int max = 0;
    for (int i = 0; i < nt * nr; i++) if (acc[i] > max) max = acc[i];
    for (int i = 0; i < nt * nr; i++)
      map[i] = (unsigned char) ((acc[i] * 255) / max);
    solution[0] = cellMax % nt;
    solution[1] = cellMax / nt;
  }
}


bool LineSpaceFilter::resize (bool larger)
{
  bool ok = false;
  if (larger)
  {
    if (nt > MIN_NB_ANG_CELLS)
    {
      nt /= 2;
      nr /= 2;
      ok = true;
    }
  }
  else
  {
    if (nt < MAX_NB_ANG_CELLS)
    {
      nt *= 2;
      nr *= 2;
      ok = true;
    }
  }
  if (ok)
  {
    delete [] acc;
    delete [] ncos;
    delete [] nsin;
    delete [] mask;
    acc = new int[nt*nr];
    ncos = new double[nt];
    nsin = new double[nt];
    mask = new bool[nt*nr];
  }
  return (ok);
}


bool LineSpaceFilter::zoom (bool in)
{
  bool ok = false;
  if (in)
  {
    if (dzoom > DIST_MIN_ZOOM) 
    {
      dzoom /= 2;
      azoom /= 2;
      ok = true;
    }
  }
  else
  {
    if (dzoom < DIST_MAX_ZOOM)
    {
      dzoom *= 2;
      azoom *= 2;
      ok = true;
    }
  }
  return (ok);
}


bool LineSpaceFilter::subpixellise (bool larger)
{
  bool ok = false;
  if (larger)
  {
    if (nbSubpix < MAX_SUBPIXEL)
    {
      nbSubpix *= 2;
      ok = true;
    }
  }
  else
  {
    if (nbSubpix > NO_SUBPIXEL)
    {
      nbSubpix /= 2;
      ok = true;
    }
  }
  return (ok);
}


bool LineSpaceFilter::setSelectivity (bool larger)
{
  bool ok = false;
  if (larger)
  {
    if (minsel < MAX_SELECTION_THRESHOLD)
    {
      minsel += INC_SELECTION_THRESHOLD;
      ok = true;
    }
  }
  else
  {
    if (minsel > MIN_SELECTION_THRESHOLD)
    {
      minsel -= INC_SELECTION_THRESHOLD;
      ok = true;
    }
  }
  return (ok);
}


vector<Pt2i> LineSpaceFilter::getAccepted ()
{
  vector<Pt2i> res;
  for (vector<Pt2i>::iterator it = leftIn.begin ();
       it != leftIn.end (); it ++)
    res.push_back (*it);
  for (vector<Pt2i>::iterator it = rightIn.begin ();
       it != rightIn.end (); it ++)
    res.push_back (*it);
  return res;
}


void LineSpaceFilter::checkManhattanScan ()
{
  if (cellMax == -1) findMax (); // default choice if not already done
  int cell0 = cellMax;
  int x0 = cell0 % nt;
  int y0 = cell0 / nt;
  int nrt = nr * nt;
  bool poursuite = true;
  int tour = 1;
int val = 1;
int vtour = 6;

  while (poursuite)
  {
    // bottom cell first
    int cell = cell0 - nt * tour;
    if (cell < 0)
    {
      // first cell on the diagonal inside the accumulator
      cell += (nt + 1) * (1 + (-cell) / nt);
      if (cell >= nt) cell = cell0;
    }
    while (cell < cell0)
    {
      // process cell HERE
if (tour == vtour) acc[cell] = val++;
      cell += nt + 1;
      if (cell % nt == 0) cell = cell0;
    }

    // right cell next
    cell = cell0 + tour;
    if (cell % nt <= x0)
    {
      // first cell on the diagonal inside the accumulator
      cell += (nt - 1) * (x0 + tour + 1 - nt);
      if (cell >= nrt) cell = cell0;
    }
    while (cell % nt > x0)
    {
      // process cell HERE
if (tour == vtour) acc[cell] = val++;
      cell += nt - 1;
      if (cell >= nrt) cell = cell0;
    }

    // top cell next
    cell = cell0 + nt * tour;
    if (cell >= nrt)
    {
      // first cell on the diagonal inside the accumulator
      cell -= (nt + 1) * (y0 + tour + 1 - nr);
      if (cell % nt >= x0) cell = cell0;
    }
    while (cell / nt > y0)
    {
      // process cell HERE
if (tour == vtour) acc[cell] = val++;
      cell -= nt + 1;
      if (cell % nt >= x0) cell = cell0;
    }

    // left cell next
    cell = cell0 - tour;
    if (tour > x0)
    {
      // first cell on the diagonal inside the accumulator
      cell -= (nt - 1) * (tour - x0);
      if (cell < 0) cell = cell0;
    }
    while (cell % nt < x0)
    {
      // process cell HERE
if (tour == vtour) acc[cell] = val++;
      cell -= nt - 1;
      if (cell < 0) cell = cell0;
    }

    tour ++;
    poursuite = (tour != 12);
  }
}


void LineSpaceFilter::init (double x0, double y0, int a, int b,
                            double length, double width)
{
  // Clears the accumulator
  int *ac = acc;
  bool *m = mask;
  for (int i = 0; i < nt * nr; i++)
  {
    *ac++ = 0;
    *m++ = false;
  }
  cellMax = -1;

  // Computes the column values (normal vector deviations)
  double w = azoom * width / nt;
  double hyp = sqrt (w * w + length * length);
  double dcos = length / hyp;
  da = acos (dcos);
  double dsin = w / hyp;
  double n = sqrt (a * a + b * b);
  ncos[nt/2] = a / n;
  nsin[nt/2] = b / n;
  for (int i = nt / 2; i > 0; i--)
  {
    ncos[i-1] = ncos[i] * dcos + nsin[i] * dsin;
    nsin[i-1] = nsin[i] * dcos - ncos[i] * dsin;
  }
  for (int i = nt / 2 + 1; i < nt; i++)
  {
    ncos[i] = ncos[i-1] * dcos - nsin[i-1] * dsin;
    nsin[i] = nsin[i-1] * dcos + ncos[i-1] * dsin;
  }

  // Computes the row values (distance parameter)
  cx = x0;
  cy = y0;
  dr = dzoom * width / nr;
  r0 = - ((nr / 2) + 0.5) * dr;
}


void LineSpaceFilter::vote (int i, int j)
{
  acc[i + j * nt] ++;
}


void LineSpaceFilter::vote (Pt2i pix)
{
  int x = pix.x () - cx;
  int y = pix.y () - cy;
  for (int i = 0; i < nt; i++)
  {
    if (nbSubpix == 1)
    {
      double d = x * ncos[i] + y * nsin[i];
      if (d >= r0)
      {
        int rho = (int) ((d - r0) / dr);
        if (rho < nr) acc[i + rho * nt] ++;
      }
    }
    else
    {
      for (int ix = 0; ix < nbSubpix; ix ++)
        for (int jy = 0; jy < nbSubpix; jy ++)
        {
          double d = (x + (2 * ix - 15) / 32.) * ncos[i]
                     + (y + (2 * jy - 15) / 32.) * nsin[i];
          if (d >= r0)
          {
            int rho = (int) ((d - r0) / dr);
            if (rho < nr) acc[i + rho * nt] ++;
          }
        }
    }
  }
}


void LineSpaceFilter::vote (const vector<Pt2i>& pts)
{
  vector<Pt2i>::const_iterator it = pts.begin ();
  while (it != pts.end ())
  {
    int x = it->x () - cx;
    int y = it->y () - cy;
    for (int i = 0; i < nt; i++)
    {
      if (nbSubpix == 1)
      {
        double d = x * ncos[i] + y * nsin[i];
        if (d >= r0)
        {
          int rho = (int) ((d - r0) / dr);
          if (rho < nr) acc[i + rho * nt] ++;
        }
      }
      else
      {
        for (int ix = 0; ix < nbSubpix; ix ++)
          for (int jy = 0; jy < nbSubpix; jy ++)
          {
            double d = (x + (2 * ix - 15) / 32.) * ncos[i]
                       + (y + (2 * jy - 15) / 32.) * nsin[i];
            if (d >= r0)
            {
              int rho = (int) ((d - r0) / dr);
              if (rho < nr) acc[i + rho * nt] ++;
            }
          }
      }
    }
    it ++;
  }
}


bool LineSpaceFilter::voteInMask (Pt2i pix)
{
  int x = pix.x () - cx;
  int y = pix.y () - cy;
  for (int i = 0; i < nt; i++)
  {
    if (nbSubpix == 1)
    {
      double d = x * ncos[i] + y * nsin[i];
      if (d >= r0)
      {
        int rho = (int) ((d - r0) / dr);
        if (rho < nr && mask[i + rho * nt]) return true;
      }
    }
    else
    {
      for (int ix = 0; ix < nbSubpix; ix ++)
        for (int jy = 0; jy < nbSubpix; jy ++)
        {
          double d = (x + (2 * ix - 15) / 32.) * ncos[i]
                     + (y + (2 * jy - 15) / 32.) * nsin[i];
          if (d >= r0)
          {
            int rho = (int) ((d - r0) / dr);
            if (rho < nr && mask[i + rho * nt]) return true;
          }
        }
    }
  }
  return false;
}


vector<Pt2i> LineSpaceFilter::voteInMask (vector<Pt2i>& pts)
{
  vector<Pt2i> vpix;
  vector<Pt2i>::iterator it = pts.begin ();
  while (it != pts.end ())
  {
    int x = it->x () - cx;
    int y = it->y () - cy;
    for (int i = 0; i < nt; i++)
    {
      double d = x * ncos[i] + y * nsin[i];
      if (d >= r0)
      {
        int rho = (int) ((d - r0) / dr);
        if (rho < nr && mask[i + rho * nt]) vpix.push_back (*it);
      }
    }
    it ++;
  }
  return (vpix);
}


void LineSpaceFilter::findMax ()
{
  cellMax = 0;
  for (int i = 1; i < nt * nr; i++)
    if (acc[i] > acc[cellMax]) cellMax = i;
}


void LineSpaceFilter::findMaxThroughOrigin ()
{
  cellMax = (nr / 2) * nt;
  for (int i = 1; i < nt; i++)
    if (acc[cellMax + i] > acc[cellMax]) cellMax = cellMax + i;
}


void LineSpaceFilter::fitMask ()
{
  if (cellMax == -1) findMax ();
  int cell0 = cellMax;
  int x0 = cell0 % nt;
  int y0 = cell0 / nt;
  int nrt = nr * nt;
  bool rt = true, tl = true, lb = true, br = true;
  int tour = 1;
  int cell;

  mask[cellMax] = true;
  while (br || rt || tl || lb)
  {
    // bottom cell first
    if (br)
    {
      br = false;
      cell = cell0 - nt * tour;
      if (cell < 0)
      {
        // first cell on the diagonal inside the accumulator
        cell += (nt + 1) * (1 + (-cell) / nt);
        if (cell >= nt) cell = cell0;
      }
      else
      {
        // process bottom cell HERE
        if (mask[cell+nt] && acc[cell] > (acc[cellMax] * minsel) / 100)
        {
          mask[cell] = true;
          br = true;
        }
        cell += nt + 1;
        if (cell % nt == 0) cell = cell0;
      }
      while (cell < cell0)
      {
        // process cell HERE
        if ((mask[cell+nt] || mask[cell-1])
            && acc[cell] > (acc[cellMax] * minsel) / 100)
        {
          mask[cell] = true;
          br = true;
        }
        cell += nt + 1;
        if (cell % nt == 0) cell = cell0;
      }
    }

    // right cell next
    if (rt)
    {
      rt = false;
      cell = cell0 + tour;
      if (cell % nt <= x0)
      {
        // first cell on the diagonal inside the accumulator
        cell += (nt - 1) * (x0 + tour + 1 - nt);
        if (cell >= nrt) cell = cell0;
      }
      else
      {
        // process left cell HERE
        if (mask[cell-1] && acc[cell] > (acc[cellMax] * minsel) / 100)
        {
          mask[cell] = true;
          rt = true;
        }
        cell += nt - 1;
        if (cell >= nrt) cell = cell0;
      }
      while (cell % nt > x0)
      {
        // process cell HERE
        if ((mask[cell-1] || mask[cell-nt])
            && acc[cell] > (acc[cellMax] * minsel) / 100)
        {
          mask[cell] = true;
          rt = true;
        }
        cell += nt - 1;
        if (cell >= nrt) cell = cell0;
      }
    }
  
    // top cell next
    if (tl)
    {
      tl = false;
      cell = cell0 + nt * tour;
      if (cell >= nrt)
      {
        // first cell on the diagonal inside the accumulator
        cell -= (nt + 1) * (y0 + tour + 1 - nr);
        if (cell % nt >= x0) cell = cell0;
      }
      else
      {
        // process top cell HERE
        if (mask[cell-nt] && acc[cell] > (acc[cellMax] * minsel) / 100)
        {
          mask[cell] = true;
          tl = true;
        }
        cell -= nt + 1;
        if (cell % nt >= x0) cell = cell0;
      }
      while (cell / nt > y0)
      {
        // process cell HERE
        if ((mask[cell-nt] || mask[cell+1])
            && acc[cell] > (acc[cellMax] * minsel) / 100)
        {
          mask[cell] = true;
          tl = true;
        }
        cell -= nt + 1;
        if (cell % nt >= x0) cell = cell0;
      }
    }

    // left cell next
    if (lb)
    {
      lb = false;
      cell = cell0 - tour;
      if (tour > x0)
      {
        // first cell on the diagonal inside the accumulator
        cell -= (nt - 1) * (tour - x0);
        if (cell < 0) cell = cell0;
      }
      else
      {
        // process left cell HERE
        if (mask[cell+1] && acc[cell] > (acc[cellMax] * minsel) / 100)
        {
          mask[cell] = true;
          lb = true;
        }
        cell -= nt - 1;
        if (cell < 0) cell = cell0;
      }
      while (cell % nt < x0)
      {
        // process cell HERE
        if ((mask[cell+1] || mask[cell+nt])
            && acc[cell] > (acc[cellMax] * minsel) / 100)
        {
          mask[cell] = true;
          lb = true;
        }
        cell -= nt - 1;
        if (cell < 0) cell = cell0;
      }
    }
    tour ++;
  }
}
