#ifndef LINE_SPACE_FILTER_H
#define LINE_SPACE_FILTER_H

#include <vector>
#include "pt2i.h"
#include "blurredsegment.h"

using namespace std;


/** 
 * @class Parameter space filter for detecting thick straight lines.
 * \author {P. Even}
 */
class LineSpaceFilter
{

public:

  /** Default number of line space filter angle cells. */
  static const int DEFAULT_NB_ANG_CELLS;
  /** Default number of line space filter distance cells. */
  static const int DEFAULT_NB_DIST_CELLS;
  

  /**
   * \brief Creates a empty line space.
   */
  LineSpaceFilter ();

  /**
   * \brief Deletes the line space.
   */
  ~LineSpaceFilter ();

  /**
   * \brief Clears the stored information.
   */
  void clear ();

  /** \brief Filters and returns a blurred segment.
    * @param bs The blurred segment to be processed.
    */
  BlurredSegment *filter (BlurredSegment *bs);

  /**
   * \brief Gets the filtering cells.
   */
  inline bool *getMask () const
  {
    return mask;
  }

  /**
   * \brief Gets the accumulator values.
   * @param map Accumulator vote cells
   * @param solution Accumulator max cell coordinates
   */
  void getAccumulator (unsigned char *map, int *solution);

  /** \brief Returns the number of columns of the accumulation filter.
    */
  inline int width () const { return nt; }

  /** \brief Returns the number of lines of the accumulation filter.
    */
  inline int height () const { return nr; }

  /** \brief Resizes the accumulation array.
    * Returns true if the required modification is actually performed.
    * @param larger Sets larger if true, smaller otherwise.
    */
  bool resize (bool larger);

  /**
   * \brief Returns the angular resolution in radians.
   */
  inline double angularResolution () const
  {
    return da;
  }

  /** \brief Returns the distance resolution in pixels.
    */
  inline double distanceResolution () const
  {
    return dr;
  }

  /** \brief Returns the distance zoom factor of the accumulator display.
    */
  inline int getZoom () const
  {
    return (dzoom);
  }

  /** \brief Zooms in or out the accumulator display.
    * Returns true if the required modification is actually performed.
    * @param in Zooms in if true, out otherwise.
    */
  bool zoom (bool in);

  /** \brief Returns the sub-pixel resolution used for voting.
    */
  inline int getSubpix () const
  {
    return (nbSubpix);
  }

  /** \brief Modifies the subpixellic resolution for voting.
    * Returns true if the required modification is actually performed.
    * @param larger Sets larger if true, smaller otherwise.
    */
  bool subpixellise (bool larger);

  /** \brief Returns the selection threshold (in percent).
    */
  inline int getSelectionThreshold () const
  {
    return (minsel);
  }

  /** \brief Modifies the filter selection threshold.
    * Returns true if the required modification is actually performed.
    * @param larger Sets larger if true, smaller otherwise.
    */
  bool setSelectivity (bool larger);

  /**
   * \brief Returns the accepted points by the line space filter.
   */
  vector<Pt2i> getAccepted ();

  /**
   * \brief Returns the rejected points by the line space filter.
   */
  inline vector<Pt2i> getRejected () const { return out; }

  /**
   * \brief Returns the blurred segment size before filtering.
   */
  inline int blurredSegmentInitialSize () const { return bsInitialSize; }

  /**
   * \brief Returns the blurred segment size after filtering.
   */
  inline int blurredSegmentFinalSize () const { return bsFinalSize; }

  /**
   * \brief Scans manhattan circles around the highest vote.
   * Votes for one specific radius to test the scan correctness.
   */
  void checkManhattanScan ();


private :

  /** Min number of line space filter angle cells. */
  static const int MIN_NB_ANG_CELLS;
  /** Max number of line space filter angle cells. */
  static const int MAX_NB_ANG_CELLS;
  /** Min number of line space filter distance cells. */
  static const int MIN_NB_DIST_CELLS;
  /** Max number of line space filter distance cells. */
  static const int MAX_NB_DIST_CELLS;
  /** Default filter selection threshold (in percentage). */
  static const int DEFAULT_SELECTION_THRESHOLD;
  /** Minimal filter selection threshold (in percentage). */
  static const int MIN_SELECTION_THRESHOLD;
  /** Maximal filter selection threshold (in percentage). */
  static const int MAX_SELECTION_THRESHOLD;
  /** Increment value for the selection threshold setting. */
  static const int INC_SELECTION_THRESHOLD;
  /** Default sub-pixellic resolution for voting. */
  static const int NO_SUBPIXEL;
  /** Maximal sub-pixellic resolution for voting. */
  static const int MAX_SUBPIXEL;
  /** Angle zoom factor default value. */
  static const int ANG_DEF_ZOOM;
  /** Angle zoom factor min value. */
  static const int ANG_MIN_ZOOM;
  /** Angle zoom factor max value. */
  static const int ANG_MAX_ZOOM;
  /** Distance zoom factor default value. */
  static const int DIST_DEF_ZOOM;
  /** Distance zoom factor min value. */
  static const int DIST_MIN_ZOOM;
  /** Distance zoom factor max value. */
  static const int DIST_MAX_ZOOM;

  /** Accumulator. */
  int *acc;
  /** Angle zoom factor. */
  int azoom; 
  /** Distance zoom factor. */
  int dzoom; 
  /** Subpixel resolution for voting. */
  int nbSubpix; 

  /** Count of accumulator rows (angles). */
  int nt;
  /** Pre-computed array of deviated normal vectors. */
  double *ncos, *nsin;
  /** Angular resolution (in readians). */
  double da;

  /** Count of accumulator columns (distances). */
  int nr;
  /** Distance resolution (offset between consecutive accumulator columns). */
  double dr;
  /** Origin of the line space. */
  int cx, cy;
  /** Distance value of the accumulator bottom line. */
  double r0;

  /** Dominant cell (-1 if not yet determined) */
  int cellMax;
  /** Extraction mask used to prune outlier candidates. */
  bool *mask;
  /** Filter selection ratio (in %) */
  int minsel;

  /** Blurred segment size before filtering. */
  int bsInitialSize;
  /** Blurred segment size after filtering. */
  int bsFinalSize;
  /** Accepted left points after line space filtering. */
  vector<Pt2i> leftIn;
  /** Accepted right points after line space filtering. */
  vector<Pt2i> rightIn;
  /** Rejected points after line space filtering. */
  vector<Pt2i> out;


  /**
   * \brief Clears the accumulator and centers it on a new blurred segment.
   * @param x0 Segment center abscissae
   * @param y0 Segment center ordinate
   * @param a Segment normal vector abscissae
   * @param b Segment normal vector ordinate
   * @param length Segment length
   * @param width Segment width
   */
  void init (double x0, double y0, int a, int b,
             double length, double width);

  /**
   * \brief Direct vote into the accumulator for testing
   */
  void vote (int i, int j);

  /**
   * \brief Adds one point votes into the accumulator
   * @param pix Voting point
   */
  void vote (Pt2i pix);

  /**
   * \brief Adds multiple points votes into the accumulator
   * @param points Voting points
   */
  void vote (const vector<Pt2i>& points);

  /**
   * \brief Checks whether the point votes in the active part of the accumulator
   * @param pix Voting point
   */
  bool voteInMask (Pt2i pix);

  /**
   * \brief Tests multiple points votes into the active part of the accumulator
   * @param points Voting points
   * Returns a vector of inlier points
   */
  vector<Pt2i> voteInMask (vector<Pt2i>& points);

  /**
   * \brief Determines the cell with the highest vote.
   */
  void findMax ();

  /**
   * \brief Determines the rho=0 cell with the highest vote.
   */
  void findMaxThroughOrigin ();

  /**
   * \brief Fits the extraction mask around the interest area.
   */
  void fitMask ();

};
#endif
